/***************************************************************************
    smb4kfileio  -  Does file IO operations for Smb4K
                             -------------------
    begin                : Do Jan 1 2004
    copyright            : (C) 2004 by Alexander Reinholdt
    email                : dustpuppy@mail.berlios.de
 ***************************************************************************/

/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,   *
 *   MA  02111-1307 USA                                                    *
 ***************************************************************************/

// Qt includes
#include <qdir.h>
#include <qfile.h>
#include <qtextstream.h>

// KDE includes
#include <kapplication.h>
#include <kdebug.h>
#include <kprocess.h>
#include <klocale.h>

// system specific includes
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <grp.h>

// application specific includes
#include "smb4kfileio.h"
#include "smb4kdefs.h"
#include "smb4kerror.h"
#include "smb4kglobal.h"
using namespace Smb4KGlobal;



Smb4KFileIO::Smb4KFileIO( QObject *parent, const char *name ) : QObject( parent, name )
{
  m_state = Idle;
  m_todo = DoNothing;
  m_error_code = None;

  connect( kapp, SIGNAL( shutDown() ), this, SLOT( slotShutdown() ) );
}


Smb4KFileIO::~Smb4KFileIO()
{
}


bool Smb4KFileIO::writeSuperUserEntries( const QString &program, bool force, bool full_use )
{
  // Do we need to do anything?
  config()->setGroup( "Super User Privileges" );
  bool program_changed = QString::compare( program, config()->readEntry( "SUID Program", "super" ) ) == 0 ? false : true;
  bool force_changed = (force == config()->readBoolEntry( "Force Unmount", false )) ? false : true;
  bool full_use_changed = (full_use == config()->readBoolEntry( "Run SUID", false )) ? false : true;

  // Stop here if nothing has changed:
  if ( !program_changed && !force_changed && !full_use_changed )
  {
    emit finished();

    return true;
  }

  bool ok = false;

  if ( QString::compare( program.stripWhiteSpace(), "super" ) == 0 )
  {
    QString filename = "super.tab";

    if ( createLockFile( filename ) )
    {
      // Check if we can read the file:
      QString canonical_path = findFile( filename );

      if ( !canonical_path.isEmpty() )
      {
        QFileInfo info( canonical_path );
        info.setCaching( false );

        if ( info.isReadable() )
        {
          QStringList contents = readFile( filename );
          bool write = true;

          // Modify the contents:
          if ( contents.grep( "# Entries for Smb4K users." ).count() == 0 )
          {
            // Make sure, we have one line distance to the previous entry:
            if ( !contents.last().stripWhiteSpace().isEmpty() )
            {
              contents.append( QString::null );
            }

            config()->setGroup( "Programs" );

            // Create the new entries:
            contents.append( "# Entries for Smb4K users." );
            contents.append( "# Generated by Smb4K. Please do not modify!" );
            contents.append( ":define Smb4KUsers "+QString( "%1" ).arg( getenv( "USER" ) ) );
#ifndef __FreeBSD__
            contents.append( "smb4k_kill\t"+config()->readPathEntry( "smb4k_kill" )+"\t$(Smb4KUsers)\tuid=root\tgid=root" );
            contents.append( "smb4k_umount\t"+config()->readPathEntry( "smb4k_umount" )+"\t$(Smb4KUsers)\tuid=root\tgid=root" );
            contents.append( "smb4k_mount\t"+config()->readPathEntry( "smb4k_mount" )+"\t$(Smb4KUsers)\tuid=root\tgid=root\tenv=PASSWD,USER" );
#else
            contents.append( "smb4k_kill\t"+config()->readPathEntry( "smb4k_kill" )+"\t$(Smb4KUsers)\tuid=root\tgid=wheel" );
            contents.append( "smb4k_umount\t"+config()->readPathEntry( "smb4k_umount" )+"\t$(Smb4KUsers)\tuid=root\tgid=wheel" );
            contents.append( "smb4k_mount\t"+config()->readPathEntry( "smb4k_mount" )+"\t$(Smb4KUsers)\tuid=root\tgid=wheel\tsetenv=HOME=$CALLER_HOME\tenv=PASSWD,USER" );
#endif
            contents.append( "# End of Smb4K user entries." );
          }
          else
          {
            // Append the user name, if necessary.
            for ( QStringList::Iterator it = contents.begin(); it != contents.end(); ++it )
            {
              if ( (*it).contains( ":define Smb4KUsers" ) != 0 )
              {
                QString users = (*it).section( "Smb4KUsers", 1, 1 ).stripWhiteSpace();
                QStringList user_list = QStringList::split( ",", users, false );

                if ( user_list.find( QString( "%1" ).arg( getenv( "USER" ) ) ) == user_list.end() )
                {
                  user_list.append( QString( "%1" ).arg( getenv( "USER" ) ) );
                  (*it).replace( users, user_list.join( "," ) );
                }
                else
                {
                  write = false;
                }

                break;
              }
              else
              {
                continue;
              }
            }
          }

          if ( write )
          {
            // Prepare the contents: remove empty lines from the end.
            QStringList::Iterator it = contents.end();

            // Move the iterator to the last entry in the list:
            --it;

            while ( (*it).stripWhiteSpace().isEmpty() )
            {
              it = contents.remove( it );
              --it;
            }

            if ( !writeFile( filename, contents ) )
            {
              Smb4KError::error( ERROR_WRITING_FILE, filename );

              emit failed();
              emit finished();

              removeLockFile();

              return ok;
            }
            else
            {
              ok = true;

              // Lock file file be removed at the end.

              emit finished();
            }
          }
          else
          {
            ok = true;

            // Lock file will be removed at the end.

            emit finished();
          }
        }
        else
        {
          // We know that the file exists, so we can get straight
          // to the reading:
          QString command;
          command.append( "kdesu -t 'smb4k_cat " );
          command.append( canonical_path+"'" );
          command.append( " ; sleep 2" );

          KProcess *p = new KProcess( this );
          p->setUseShell( true );

          connect( p,   SIGNAL( receivedStderr( KProcess *, char *, int ) ),
                  this, SLOT( slotReceivedStderr( KProcess *, char *, int ) ) );
          connect( p,   SIGNAL( receivedStdout( KProcess *, char *, int ) ),
                  this, SLOT( slotReceivedStdout( KProcess *, char *, int ) ) );
          connect( p,   SIGNAL( processExited( KProcess * ) ),
                  this, SLOT( slotProcessExited( KProcess * ) ) );

          *p << command;

          m_state = ReadSuperTab;
          m_todo = SuperTabAdd;

          p->start( KProcess::NotifyOnExit, KProcess::AllOutput );
        }
      }
      else
      {
        Smb4KError::error( ERROR_FILE_NOT_FOUND, filename );

        emit failed();
        emit finished();

        removeLockFile();

        return ok;
      }
    }
    else
    {
      // The error message will be shown by Smb4KFileIO::createLockFile()
      emit failed();
      emit finished();

      // We must not remove the lock file here.

      return ok;
    }
  }
  else if ( QString::compare( program.stripWhiteSpace(), "sudo" ) == 0 )
  {
    QString filename = "sudoers";

    if ( createLockFile( filename ) )
    {
      // Check if we can read the file:
      QString canonical_path = findFile( filename );

      if ( !canonical_path.isEmpty() )
      {
        QFileInfo info( canonical_path );
        info.setCaching( false );

        if ( info.isReadable() )
        {
          QStringList contents = readFile( filename );
          bool write = false;

          size_t hostnamelen = 255;
          char *hn = new char[hostnamelen];

          if ( gethostname( hn, hostnamelen ) == -1 )
          {
            int error = errno;
            Smb4KError::error( ERROR_GETTING_HOSTNAME, QString::null, strerror( error ) );

            removeLockFile();

            return ok; // false
          }

          QString hostname( hn );

          delete [] hn;

          if ( contents.grep( "# Entries for Smb4K users." ).count() == 0 )
          {
            config()->setGroup( "Programs" );

            contents.append( "# Entries for Smb4K users." );
            contents.append( "# Generated by Smb4K. Please do not modify!" );
            contents.append( "User_Alias\tSMB4KUSERS = "+QString( "%1" ).arg( getenv( "USER" ) ) );
            contents.append( "Defaults:SMB4KUSERS\tenv_keep=\"PASSWD USER\"" );
            contents.append( "SMB4KUSERS\t"+hostname+" = NOPASSWD: "+config()->readPathEntry( "smb4k_kill" ) );
            contents.append( "SMB4KUSERS\t"+hostname+" = NOPASSWD: "+config()->readPathEntry( "smb4k_umount" ) );
            contents.append( "SMB4KUSERS\t"+hostname+" = NOPASSWD: "+config()->readPathEntry( "smb4k_mount" ) );
            contents.append( "# End of Smb4K user entries." );

            write = true;
          }
          else
          {
            for ( QStringList::Iterator it = contents.begin(); it != contents.end(); ++it )
            {
              if ( (*it).startsWith( "User_Alias\tSMB4KUSERS" ) && (*it).contains( getenv( "USER" ), true ) == 0 )
              {
                (*it).append( ","+QString( getenv( "USER" ) ) );

                write = true;

                continue;
              }
              else if ( (*it).startsWith( "Defaults:SMB4KUSERS" ) && (*it).contains( " USER", true ) == 0 )
              {
                *it = "Defaults:SMB4KUSERS\tenv_keep=\"PASSWD USER\"";

                write = true;

                continue;
              }
              else if ( (*it).startsWith( "# End of Smb4K user entries." ) )
              {
                break;
              }
              else
              {
                continue;
              }
            }
          }

          if ( write )
          {
            // Prepare the contents: remove empty lines from the end.
            QStringList::Iterator it = contents.end();

            // Move the iterator to the last entry in the list:
            --it;

            while ( (*it).stripWhiteSpace().isEmpty() )
            {
              it = contents.remove( it );
              --it;
            }

            if ( !writeFile( filename, contents ) )
            {
              Smb4KError::error( ERROR_WRITING_FILE, filename );

              emit failed();
              emit finished();

              removeLockFile();

              return ok; // false
            }
            else
            {
              ok = true;

              // Lock file file be removed at the end.

              emit finished();
            }
          }
          else
          {
            ok = true;

            // Lock file file be removed at the end.

            emit finished();
          }
        }
        else
        {
          // We know that the file exists, so we can get straight
          // to the reading:
          QString command;
          command.append( "kdesu -t 'smb4k_cat " );
          command.append( canonical_path+"'" );
          command.append( " ; sleep 2" );

          KProcess *p = new KProcess( this );
          p->setUseShell( true );

          connect( p,   SIGNAL( receivedStderr( KProcess *, char *, int ) ),
                  this, SLOT( slotReceivedStderr( KProcess *, char *, int ) ) );
          connect( p,   SIGNAL( receivedStdout( KProcess *, char *, int ) ),
                  this, SLOT( slotReceivedStdout( KProcess *, char *, int ) ) );
          connect( p,   SIGNAL( processExited( KProcess * ) ),
                  this, SLOT( slotProcessExited( KProcess * ) ) );

          *p << command;

          m_state = ReadSudoers;
          m_todo = SudoersAdd;

          p->start( KProcess::NotifyOnExit, KProcess::AllOutput );
        }
      }
      else
      {
        Smb4KError::error( ERROR_FILE_NOT_FOUND, filename );

        emit failed();
        emit finished();

        removeLockFile();

        return ok; // false
      }
    }
    else
    {
      // The error message will be shown by Smb4KFileIO::createLockFile()
      emit failed();
      emit finished();

      // We must not remove the lock file here.

      return ok;
    }
  }
  else
  {
    // Do nothing here.
  }

  ok = removeLockFile();

  return ok;
}


bool Smb4KFileIO::removeSuperUserEntries( const QString &program )
{
  // First thing is to set the appropriate config entries
  // to FALSE:
  config()->setGroup( "Super User Privileges" );
  config()->writeEntry( "Force Unmount", false );
  config()->writeEntry( "Run SUID", false );
  config()->sync();

  // Now modify the sudoers or super.tab file:
  bool ok = false;

  if ( QString::compare( "super", program ) == 0 )
  {
    QString filename = "super.tab";

    if ( createLockFile( filename ) )
    {
      QString canonical_path = findFile( filename );

      if ( !canonical_path.isEmpty() )
      {
        QFileInfo info( canonical_path );
        info.setCaching( false );

        if ( info.isReadable() )
        {
          QStringList contents = readFile( filename );
          bool write = false;

          // Find the beginning and the end of the entries in
          // the super.tab file:
          QStringList::Iterator begin = contents.find( "# Entries for Smb4K users." );
          QStringList::Iterator end = contents.find( "# End of Smb4K user entries." );

          // Now, check if the user is in the list of users. If he is,
          // remove him from there or remove all if he is the only one:
          for ( QStringList::Iterator it = begin; it != end; ++it )
          {
            if ( (*it).startsWith( ":define Smb4KUsers" ) )
            {
              QString users = (*it).section( "Smb4KUsers", 1, 1 ).stripWhiteSpace();

              if ( users.contains( "," ) == 0 )
              {
                // In this case, there is only one user in the list. Check if
                // it is the user who requested the removal:
                if ( QString::compare( users, getenv( "USER" ) ) == 0 )
                {
                  // They are equal. Remove everything:
                  contents.erase( begin, end );
                  contents.remove( end );

                  write = true;

                  break;
                }
                else
                {
                  // They are not equal: Do nothing.
                  break;
                }
              }
              else
              {
                // In this case there are more than one user in the list.
                // Remove the user who requested the removal from the list:
                QStringList list = QStringList::split( ",", users, false );
                list.remove( getenv( "USER" ) );

                (*it).replace( users, list.join( "," ) );

                write = true;

                break;
              }
            }
            else
            {
              continue;
            }
          }

          if ( write )
          {
            // Prepare the contents: remove empty lines from the end.
            QStringList::Iterator it = contents.end();

            // Move the iterator to the last entry in the list:
            --it;

            while ( (*it).stripWhiteSpace().isEmpty() )
            {
              it = contents.remove( it );
              --it;
            }

            if ( !writeFile( filename, contents ) )
            {
              Smb4KError::error( ERROR_WRITING_FILE, filename );

              emit failed();
              emit finished();

              removeLockFile();

              return ok;
            }
            else
            {
              ok = true;

              // Lock file file be removed at the end.

              emit finished();
            }
          }
          else
          {
            ok = true;

            // Lock file will be removed at the end.

            emit finished();
          }
        }
        else
        {
          // We know that the file exists, so we can get straight
          // to the reading:
          QString command;
          command.append( "kdesu -t 'smb4k_cat " );
          command.append( canonical_path+"'" );
          command.append( " ; sleep 2" );

          KProcess *p = new KProcess( this );
          p->setUseShell( true );

          connect( p,   SIGNAL( receivedStderr( KProcess *, char *, int ) ),
                  this, SLOT( slotReceivedStderr( KProcess *, char *, int ) ) );
          connect( p,   SIGNAL( receivedStdout( KProcess *, char *, int ) ),
                  this, SLOT( slotReceivedStdout( KProcess *, char *, int ) ) );
          connect( p,   SIGNAL( processExited( KProcess * ) ),
                  this, SLOT( slotProcessExited( KProcess * ) ) );

          *p << command;

          m_state = ReadSuperTab;
          m_todo = SuperTabRemove;

          p->start( KProcess::NotifyOnExit, KProcess::AllOutput );
        }
      }
      else
      {
        Smb4KError::error( ERROR_FILE_NOT_FOUND, filename );

        emit failed();
        emit finished();

        removeLockFile();

        return ok; // false
      }
    }
    else
    {
      // The error message will be shown by Smb4KFileIO::createLockFile()
      emit failed();
      emit finished();

      // We must not remove the lock file here.

      return ok;
    }
  }
  else if ( QString::compare( "sudo", program ) == 0 )
  {
    QString filename = "sudoers";

    if ( createLockFile( filename ) )
    {
      QString canonical_path = findFile( filename );

      if ( !canonical_path.isEmpty() )
      {
        QFileInfo info( canonical_path );
        info.setCaching( false );

        if ( info.isReadable() )
        {
          QStringList contents = readFile( filename );
          bool write = false;

          // Find the beginning and the end of the entries in
          // the super.tab file:
          QStringList::Iterator begin = contents.find( "# Entries for Smb4K users." );
          QStringList::Iterator end = contents.find( "# End of Smb4K user entries." );

          // Now, check if the user is in the list of users. If he is,
          // remove him from there or remove all if he is the only one:
          for ( QStringList::Iterator it = begin; it != end; ++it )
          {
            if ( (*it).startsWith( "User_Alias\tSMB4KUSERS" ) )
            {
              QString users = (*it).section( "=", 1, 1 ).stripWhiteSpace();

              if ( users.contains( "," ) == 0 )
              {
                // In this case, there is only one user in the list. Check if
                // it is the user who requested the removal:
                if ( QString::compare( users, getenv( "USER" ) ) == 0 )
                {
                  // They are equal. Remove everything:
                  contents.erase( begin, end );
                  contents.remove( end );

                  write = true;

                  break;
                }
                else
                {
                  // They are not equal: Do nothing.
                  break;
                }
              }
              else
              {
                // In this case there are more than one user in the list.
                // Remove the user who requested the removal from the list:
                QStringList list = QStringList::split( ",", users, false );
                list.remove( getenv( "USER" ) );

                (*it).replace( users, list.join( "," ) );

                write = true;

                break;
              }
            }
            else
            {
              continue;
            }
          }

          if ( write )
          {
            // Prepare the contents: remove empty lines from the end.
            QStringList::Iterator it = contents.end();

            // Move the iterator to the last entry in the list:
            --it;

            while ( (*it).stripWhiteSpace().isEmpty() )
            {
              it = contents.remove( it );
              --it;
            }

            if ( !writeFile( filename, contents ) )
            {
              Smb4KError::error( ERROR_WRITING_FILE, filename );

              emit failed();
              emit finished();

              removeLockFile();

              return ok;
            }
            else
            {
              ok = true;

              // Lock file file be removed at the end.

              emit finished();
            }
          }
          else
          {
            ok = true;

            // Lock file will be removed at the end.

            emit finished();
          }
        }
        else
        {
          // We know that the file exists, so we can get straight
          // to the reading:
          QString command;
          command.append( "kdesu -t 'smb4k_cat " );
          command.append( canonical_path+"'" );
          command.append( " ; sleep 2" );

          KProcess *p = new KProcess( this );
          p->setUseShell( true );

          connect( p,   SIGNAL( receivedStderr( KProcess *, char *, int ) ),
                  this, SLOT( slotReceivedStderr( KProcess *, char *, int ) ) );
          connect( p,   SIGNAL( receivedStdout( KProcess *, char *, int ) ),
                  this, SLOT( slotReceivedStdout( KProcess *, char *, int ) ) );
          connect( p,   SIGNAL( processExited( KProcess * ) ),
                  this, SLOT( slotProcessExited( KProcess * ) ) );

          *p << command;

          m_state = ReadSudoers;
          m_todo = SudoersRemove;

          p->start( KProcess::NotifyOnExit, KProcess::AllOutput );
        }
      }
      else
      {
        Smb4KError::error( ERROR_FILE_NOT_FOUND, filename );

        emit failed();
        emit finished();

        removeLockFile();

        return ok; // false
      }
    }
    else
    {
      // The error message will be shown by Smb4KFileIO::createLockFile()
      emit failed();
      emit finished();

      // We must not remove the lock file here.

      return ok;
    }
  }
  else
  {
    // Do nothing
  }

  ok = removeLockFile();

  return ok;
}


const QValueList<Smb4KUser *> Smb4KFileIO::getUsers()
{
  QStringList contents = readFile( "passwd" );

  QValueList<Smb4KUser *> users;

  for ( QStringList::ConstIterator it = contents.begin(); it != contents.end(); ++it )
  {
    if ( !(*it).stripWhiteSpace().startsWith( "#" ) &&
         ((*it).section( ":", 2, 2 ).toInt() >= 500 /* 500 is the lowerst known user ID */ ||
         (*it).section( ":", 2, 2 ).toInt() == (int)getuid()) )
    {
      users.append( new Smb4KUser( (*it).section( ":", 2, 2 ).toInt(), (*it).section( ":", 3, 3 ).toInt() ) );

      continue;
    }
    else
    {
      continue;
    }
  }

  return users;
}


QStringList Smb4KFileIO::readFile( const QString &filename )
{
  QString canonical_path = findFile( filename );

  QStringList contents;

  if ( !canonical_path.isEmpty() )
  {
    QFile file( canonical_path );

    if ( file.open( IO_ReadOnly ) )
    {
      QTextStream ts( &file );
      ts.setEncoding( QTextStream::Locale );

      contents = QStringList::split( '\n', ts.read(), true );

      file.close();
    }
    else
    {
      Smb4KError::error( ERROR_READING_FILE, filename );

      // empty:
      return contents;
    }
  }
  else
  {
    Smb4KError::error( ERROR_FILE_NOT_FOUND, filename );

    // empty:
    return contents;
  }

  return contents;
}


bool Smb4KFileIO::writeFile( const QString &filename, const QStringList &contents )
{
  bool ok = false;

  QString canonical_path = findFile( filename );

  if ( !canonical_path.isEmpty() )
  {
    QFileInfo info( canonical_path );
    info.setCaching( false );

    if ( info.isWritable() )
    {
      QFile file( canonical_path );

      // Permissions will be preserved by this action!
      if ( file.open( IO_WriteOnly ) )
      {
        QTextStream ts( &file );
        ts.setEncoding( QTextStream::Locale );

        ts << contents.join( "\n" ) << endl;

        file.close();
      }
      else
      {
        Smb4KError::error( ERROR_WRITING_FILE, file.name() );

        return ok; // is false
      }
    }
    else
    {
      // Create a temporary file and write the data to it:
      QCString template_string = tempDir().local8Bit()+"/XXXXXX";
      char tmp[template_string.length()+1];
      (void) qstrncpy( tmp, template_string, template_string.length()+1 );

      QFile temp_file;
      int file_descriptor;

      if ( (file_descriptor = mkstemp( tmp )) == -1 )
      {
        int err = errno;

        Smb4KError::error( ERROR_CREATING_TEMP_FILE, tmp, strerror( err ) );

        return ok; // still false
      }

      if ( temp_file.open( IO_WriteOnly, file_descriptor ) )
      {
        QTextStream ts( &temp_file );
        ts.setEncoding( QTextStream::Locale );

        ts << contents.join( "\n" ) << endl;

        temp_file.close();
      }
      else
      {
        Smb4KError::error( ERROR_WRITING_FILE, temp_file.name() );

        return ok; // still false
      }

      // Now move the file to the right location. Preserve the permissions
      // and the owner:
      struct stat file_stat;

      if ( stat( canonical_path.local8Bit(), &file_stat ) == -1 )
      {
        int error_number = errno;

        Smb4KError::error( ERROR_GETTING_PERMISSIONS, QString::null, strerror( error_number ) );

        return ok; // still false
      }

      QString perms = QString( "%1" ).arg( (int)file_stat.st_mode, 0, 8 );
      perms = perms.right( 4 );
      QString owner = QString( "%1" ).arg( (int)file_stat.st_uid );
      QString group = QString( "%1" ).arg( (int)file_stat.st_gid );
      QString temp_file_name = QString( tmp );

      // Assemble the command.
      QString command;
      command.append( "kdesu -n 'smb4k_mv "+owner+":"+group+" "+perms+" " );
      command.append( temp_file_name+" " );
      command.append( canonical_path+"' ; " );
      command.append( "rm -f "+temp_file_name );

      // Process the command:
      KProcess *p = new KProcess( this );
      p->setUseShell( true );

      connect( p,   SIGNAL( receivedStderr( KProcess *, char *, int ) ),
              this, SLOT( slotReceivedStderr( KProcess *, char *, int ) ) );
      connect( p,   SIGNAL( receivedStdout( KProcess *, char *, int ) ),
              this, SLOT( slotReceivedStdout( KProcess *, char *, int ) ) );
      connect( p,   SIGNAL( processExited( KProcess * ) ),
              this, SLOT( slotProcessExited( KProcess * ) ) );

      if ( QString::compare( filename, "super.tab" ) == 0 )
      {
        m_state = WriteSuperTab;
      }
      else if ( QString::compare( filename, "sudoers" ) == 0 )
      {
        m_state = WriteSudoers;
      }

      *p << command;

      p->start( KProcess::NotifyOnExit, KProcess::AllOutput );
    }

    ok = true;
  }
  else
  {
    Smb4KError::error( ERROR_FILE_NOT_FOUND );

    return ok;
  }

  return ok;
}


bool Smb4KFileIO::createLockFile( const QString &filename )
{
  bool ok = false;

  // Determine the directory where to write the lock file. First, try
  // /var/lock and than /var/tmp. If that does not work either, fall
  // back to /tmp.
  if ( m_lock_file.isEmpty() )
  {
    QValueList<QCString> dirs;
    dirs << "/var/lock" << "/var/tmp" << "/tmp";

    struct stat buf;

    for ( QValueList<QCString>::ConstIterator it = dirs.begin(); it != dirs.end(); ++it )
    {
      // First check if /var/lock is available and writable
      if ( lstat( *it, &buf ) == -1 )
      {
        int error_number = errno;

        if ( error_number != EACCES && error_number != ENOENT )
        {
          Smb4KError::error( ERROR_GETTING_PERMISSIONS, *it, strerror( error_number ) );

          return ok; // false
        }
      }
      else
      {
        // Look for the groups the user is in:
        long ngroups_max;
        ngroups_max = sysconf(_SC_NGROUPS_MAX);

        gid_t list[ngroups_max];

        if ( getgroups( ngroups_max, list ) == -1 )
        {
          int error_number = errno;

          Smb4KError::error( ERROR_GETTING_GIDS, QString::null, strerror( error_number ) );

          return ok; // false
        }

        gid_t sup_gid = 65534; // set this to gid 'nobody' for initialization
        bool found_gid = false;
        int i = 0;

        while ( list[i] )
        {
          if ( list[i] == buf.st_gid )
          {
            sup_gid = list[i];
            found_gid = true;
          }

          i++;
        }

        // Check whether we are stat'ing a directory and that the
        // user has read/write permissions.
        if ( S_ISDIR( buf.st_mode ) /* is directory */ &&
            (buf.st_uid == getuid() && (buf.st_mode & 00600) == (S_IWUSR | S_IRUSR)) /* user */ ||
            (found_gid && buf.st_gid == sup_gid && (buf.st_mode & 00060) == (S_IWGRP | S_IRGRP)) /* group */ ||
            ((buf.st_mode & 00006) == (S_IWOTH | S_IROTH)) /* others */ )
        {
          m_lock_file = *it+"/smb4k.lock";

          break;
        }
        else
        {
          continue;
        }
      }
    }
  }

  int file_descriptor;

  // Create the lock file if necessary and open it:
  if ( (file_descriptor = open( m_lock_file, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH )) == -1 )
  {
    // Error out if the opening failed:
    int error_number = errno;

    Smb4KError::error( ERROR_OPENING_FILE, m_lock_file, strerror( error_number ) );

    return ok; // false
  }
  else
  {
    // Check what we actually opened:

    struct stat file_stat;

    if ( fstat( file_descriptor, &file_stat ) == -1 )
    {
      // Error out if we could not get the information about the file:
      int error_number = errno;

      // FIXME for >= 0.8.x: Change error code to ERROR_GETTING_STAT
      Smb4KError::error( ERROR_GETTING_PERMISSIONS, QString::null, strerror( error_number ) );

      return ok;
    }

    if ( !S_ISREG( file_stat.st_mode ) || S_ISFIFO( file_stat.st_mode ) || S_ISLNK( file_stat.st_mode ) )
    {
      // Close the file and error out, if we have opened an
      // "irregular" file (i.e. a symlink, a fifo, etc.).

      Smb4KError::error( ERROR_LOCK_FILE_IS_SYMLINK, m_lock_file );

      if ( close( file_descriptor ) == -1 )
      {
        int error_number = errno;

        // FIXME for >= 0.8.x: Change error code to ERROR_CLOSING_FILE
        Smb4KError::error( ERROR_UNKNOWN, QString::null, strerror( error_number ) );
      }

      return ok; // false
    }
    else
    {
      // Continue if the file is regular.

      char buffer[1000];
      ssize_t size;

      if ( (size = read( file_descriptor, buffer, 1000 )) == -1 )
      {
        int error_number = errno;

        Smb4KError::error( ERROR_READING_FILE, m_lock_file, strerror( error_number ) );

        return ok; // false
      }

      if ( size >= 1000 )
      {
        // FIXME for >= 0.8.x: Change error code to ERROR_BUFFER_EXCEEDED
        Smb4KError::error( ERROR_UNKNOWN, QString::null, i18n( "Buffer size exceeded" ) );

        return ok; // false
      }

      QStringList contents = QStringList::split( '\n', QString::fromLocal8Bit( buffer, size ), false );
      QString test_string = ":"+filename;
      QString entry = contents.grep( test_string, true ).join( "\n" ).stripWhiteSpace();

      if ( !entry.isEmpty() )
      {
        Smb4KError::error( ERROR_LOCKED, entry );

        return ok; // false
      }
      else
      {
        contents << QString( "%1:%2" ).arg( getenv( "USER" ) ).arg( filename );
        QCString out = contents.join( "\n" ).local8Bit();

        if ( write( file_descriptor, out, out.length() ) == -1 )
        {
          int error_number = errno;

          Smb4KError::error( ERROR_WRITING_FILE, m_lock_file, strerror( error_number ) );

          return ok; // false
        }

        if ( close( file_descriptor ) == -1 )
        {
          int error_number = errno;

          // FIXME for >= 0.8.x: Change error code to ERROR_CLOSING_FILE
          Smb4KError::error( ERROR_UNKNOWN, QString::null, strerror( error_number ) );

          return ok; // false
        }

        ok = true;
      }
    }
  }

  return ok;
}


bool Smb4KFileIO::removeLockFile( const bool error_message )
{
  // We already have the name and location of the lock
  // file, so we do not need to define it here.

  int file_descriptor;
  bool ok = false;

  // Open the lock file:
  if ( (file_descriptor = open( m_lock_file, O_RDWR )) == -1 )
  {
    // Error out if the opening failed:
    int error_number = errno;

    if ( error_message && error_number != ENOENT )
    {
      Smb4KError::error( ERROR_OPENING_FILE, m_lock_file, strerror( error_number ) );
    }

    return ok; // false
  }
  else
  {
    // Check what we actually opened:

    struct stat file_stat;

    if ( fstat( file_descriptor, &file_stat ) == -1 )
    {
      // Error out if we could not get the information about the file:
      int error_number = errno;

      if ( error_message )
      {
        // FIXME for >= 0.8.x: Change error code to ERROR_GETTING_STAT
        Smb4KError::error( ERROR_GETTING_PERMISSIONS, QString::null, strerror( error_number ) );
      }

      return ok;
    }

    if ( !S_ISREG( file_stat.st_mode ) || S_ISFIFO( file_stat.st_mode ) || S_ISLNK( file_stat.st_mode ) )
    {
      // Close the file and error out, if we have opened an
      // "irregular" file (i.e. a symlink, a fifo, etc.).

      if ( error_message )
      {
        Smb4KError::error( ERROR_LOCK_FILE_IS_SYMLINK, m_lock_file );
      }

      if ( close( file_descriptor ) == -1 )
      {
        int error_number = errno;

        if ( error_message )
        {
          // FIXME for >= 0.8.x: Change error code to ERROR_CLOSING_FILE
          Smb4KError::error( ERROR_UNKNOWN, QString::null, strerror( error_number ) );
        }
      }

      return ok; // false
    }
    else
    {
      // Continue if the file is regular.

      char buffer[1000];
      ssize_t size;

      if ( (size = read( file_descriptor, buffer, 1000 )) == -1 )
      {
        int error_number = errno;

        if ( error_message )
        {
          Smb4KError::error( ERROR_READING_FILE, m_lock_file, strerror( error_number ) );
        }

        return ok; // false
      }

      if ( size >= 1000 )
      {
        if ( error_message )
        {
          // FIXME for >= 0.8.x: Change error code to ERROR_BUFFER_EXCEEDED
          Smb4KError::error( ERROR_UNKNOWN, QString::null, i18n( "Buffer size exceeded" ) );
        }

        return ok; // false
      }

      QStringList contents = QStringList::split( '\n', QString::fromLocal8Bit( buffer, size ), false );

      // Prepare the contents of the file and write it to the disk.
      // It it should be empty, remove the lock file.
      for ( QStringList::Iterator it = contents.begin(); it != contents.end(); it++ )
      {
        if ( (*it).startsWith( QString( getenv( "USER" ) )+":" ) )
        {
          *it = QString::null;

          continue;
        }
        else
        {
          continue;
        }
      }

      contents.remove( QString::null );

      if ( !contents.isEmpty() )
      {
        // Write the remaining contents to the lock file:

        QCString out = contents.join( "\n" ).local8Bit();

        if ( write( file_descriptor, out, out.length() ) == -1 )
        {
          int error_number = errno;

          if ( error_message )
          {
            Smb4KError::error( ERROR_WRITING_FILE, m_lock_file, strerror( error_number ) );
          }

          return ok; // false
        }

        if ( close( file_descriptor ) == -1 )
        {
          int error_number = errno;

          if ( error_message )
          {
            // FIXME for > 0.8.x: Change error code to ERROR_CLOSING_FILE
            Smb4KError::error( ERROR_UNKNOWN, QString::null, strerror( error_number ) );
          }

          return ok; // false
        }

        ok = true;
      }
      else
      {
        // Close and remove the lock file:

        if ( close( file_descriptor ) == -1 )
        {
          int error_number = errno;

          if ( error_message )
          {
            // FIXME for > 0.8.x: Change error code to ERROR_CLOSING_FILE
            Smb4KError::error( ERROR_UNKNOWN, QString::null, strerror( error_number ) );
          }

          return ok; // false
        }

        if ( unlink( m_lock_file ) == -1 )
        {
          int error_number = errno;

          if ( error_message )
          {
            // FIXME for > 0.8.x: Replace error code with ERROR_REMOVING_FILE
            Smb4KError::error( ERROR_UNKNOWN, m_lock_file, strerror( error_number ) );
          }

          return ok; // false
        }

        ok = true;
      }
    }
  }

  return ok;
}


const QString Smb4KFileIO::findFile( const QString &filename )
{
  QStringList paths;
  paths << "/etc";
  paths << "/etc/samba";
  paths << "/usr/local/etc";
  paths << "/usr/local/etc/samba";

  QString canonical_path = QString::null;

  for ( QStringList::ConstIterator it = paths.begin(); it != paths.end(); it++ )
  {
    QDir::setCurrent( *it );

    if ( QFile::exists( filename ) )
    {
      canonical_path = QDir::current().canonicalPath()+"/"+filename;

      break;
    }
    else
    {
      continue;
    }
  }

  return canonical_path;
}


void Smb4KFileIO::processOutput( const QString &filename )
{
  // Stop here if the user cancelled the process, i.e. if
  // the buffer is empty.
  if ( m_output_buffer.isEmpty() )
  {
    emit failed();
    emit finished();

    removeLockFile();

    return;
  }

  QStringList contents = QStringList::split( '\n', m_output_buffer, true );

  if ( m_todo == SuperTabAdd )
  {
    bool write = false;

    if ( contents.grep( "# Entries for Smb4K users." ).count() == 0 )
    {
      // Make sure, we have one line distance to the previous entry:
      if ( !contents.last().stripWhiteSpace().isEmpty() )
      {
        contents.append( QString::null );
      }

      config()->setGroup( "Programs" );

      // Create the new entries:
      contents.append( "# Entries for Smb4K users." );
      contents.append( "# Generated by Smb4K. Please do not modify!" );
      contents.append( ":define Smb4KUsers "+QString( "%1" ).arg( getenv( "USER" ) ) );
#ifndef __FreeBSD__
      contents.append( "smb4k_kill\t"+config()->readPathEntry( "smb4k_kill" )+"\t$(Smb4KUsers)\tuid=root\tgid=root" );
      contents.append( "smb4k_umount\t"+config()->readPathEntry( "smb4k_umount" )+"\t$(Smb4KUsers)\tuid=root\tgid=root" );
      contents.append( "smb4k_mount\t"+config()->readPathEntry( "smb4k_mount" )+"\t$(Smb4KUsers)\tuid=root\tgid=root\tenv=PASSWD,USER" );
#else
      contents.append( "smb4k_kill\t"+config()->readPathEntry( "smb4k_kill" )+"\t$(Smb4KUsers)\tuid=root\tgid=wheel" );
      contents.append( "smb4k_umount\t"+config()->readPathEntry( "smb4k_umount" )+"\t$(Smb4KUsers)\tuid=root\tgid=wheel" );
      contents.append( "smb4k_mount\t"+config()->readPathEntry( "smb4k_mount" )+"\t$(Smb4KUsers)\tuid=root\tgid=wheel\tsetenv=HOME=$CALLER_HOME\tenv=PASSWD,USER" );
#endif
      contents.append( "# End of Smb4K user entries." );

      write = true;
    }
    else
    {
      // Append the user name, if necessary.
      for ( QStringList::Iterator it = contents.begin(); it != contents.end(); ++it )
      {
        if ( (*it).contains( ":define Smb4KUsers" ) != 0 )
        {
          QString users = (*it).section( "Smb4KUsers", 1, 1 ).stripWhiteSpace();
          QStringList user_list = QStringList::split( ",", users, false );

          if ( user_list.find( QString( "%1" ).arg( getenv( "USER" ) ) ) == user_list.end() )
          {
            user_list.append( QString( "%1" ).arg( getenv( "USER" ) ) );
            (*it).replace( users, user_list.join( "," ) );

            write = true;
          }

          break;
        }
        else
        {
          continue;
        }
      }
    }

    if ( write )
    {
      // Prepare the contents: remove empty lines from the end.
      QStringList::Iterator it = contents.end();

      // Move the iterator to the last entry in the list:
      --it;

      while ( (*it).stripWhiteSpace().isEmpty() )
      {
        it = contents.remove( it );
        --it;
      }

      if ( !writeFile( filename, contents ) )
      {
        Smb4KError::error( ERROR_WRITING_FILE, filename );

        emit failed();
        emit finished();
      }
      else
      {
        emit finished();
      }
    }
    else
    {
      emit finished();
    }
  }
  else if ( m_todo == SudoersAdd )
  {
    bool write = false;

    size_t hostnamelen = 255;
    char *hn = new char[hostnamelen];

    if ( gethostname( hn, hostnamelen ) == -1 )
    {
      int error = errno;
      Smb4KError::error( ERROR_GETTING_HOSTNAME, QString::null, strerror( error ) );

      removeLockFile();
    }

    QString hostname( hn );

    delete [] hn;

    if ( contents.grep( "# Entries for Smb4K users." ).count() == 0 )
    {
      config()->setGroup( "Programs" );

      contents.append( "# Entries for Smb4K users." );
      contents.append( "# Generated by Smb4K. Please do not modify!" );
      contents.append( "User_Alias\tSMB4KUSERS = "+QString( "%1" ).arg( getenv( "USER" ) ) );
      contents.append( "Defaults:SMB4KUSERS\tenv_keep=\"PASSWD USER\"" );
      contents.append( "SMB4KUSERS\t"+hostname+" = NOPASSWD: "+config()->readPathEntry( "smb4k_kill" ) );
      contents.append( "SMB4KUSERS\t"+hostname+" = NOPASSWD: "+config()->readPathEntry( "smb4k_umount" ) );
      contents.append( "SMB4KUSERS\t"+hostname+" = NOPASSWD: "+config()->readPathEntry( "smb4k_mount" ) );
      contents.append( "# End of Smb4K user entries." );

      write = true;
    }
    else
    {
      for ( QStringList::Iterator it = contents.begin(); it != contents.end(); ++it )
      {
        if ( (*it).startsWith( "User_Alias\tSMB4KUSERS" ) && (*it).contains( getenv( "USER" ), true ) == 0 )
        {
          (*it).append( ","+QString( getenv( "USER" ) ) );

          write = true;

          continue;
        }
        else if ( (*it).startsWith( "Defaults:SMB4KUSERS" ) && (*it).contains( " USER", true ) == 0 )
        {
          *it = "Defaults:SMB4KUSERS\tenv_keep=\"PASSWD USER\"";

          write = true;

          continue;
        }
        else if ( (*it).startsWith( "# End of Smb4K user entries." ) )
        {
          break;
        }
        else
        {
          continue;
        }
      }
    }

    if ( write )
    {
      // Prepare the contents: remove empty lines from the end.
      QStringList::Iterator it = contents.end();

      // Move the iterator to the last entry in the list:
      --it;

      while ( (*it).stripWhiteSpace().isEmpty() )
      {
        it = contents.remove( it );
        --it;
      }

      if ( !writeFile( filename, contents ) )
      {
        Smb4KError::error( ERROR_WRITING_FILE, filename );

        emit failed();
        emit finished();
      }
      else
      {
        emit finished();
      }
    }
    else
    {
      emit finished();
    }
  }
  else if ( m_todo == SuperTabRemove )
  {
    bool write = false;

    // Find the beginning and the end of the entries in
    // the super.tab file:
    QStringList::Iterator begin = contents.find( "# Entries for Smb4K users." );
    QStringList::Iterator end = contents.find( "# End of Smb4K user entries." );

    // Now, check if the user is in the list of users. If he is,
    // remove him from there or remove all if he is the only one:
    for ( QStringList::Iterator it = begin; it != end; ++it )
    {
      if ( (*it).startsWith( ":define Smb4KUsers" ) )
      {
        QString users = (*it).section( "Smb4KUsers", 1, 1 ).stripWhiteSpace();

        if ( users.contains( "," ) == 0 )
        {
          // In this case, there is only one user in the list. Check if
          // it is the user who requested the removal:
          if ( QString::compare( users, getenv( "USER" ) ) == 0 )
          {
            // They are equal. Remove everything:
            contents.erase( begin, end );
            contents.remove( end );

            write = true;

            break;
          }
          else
          {
            // They are not equal: Do nothing.
            break;
          }
        }
        else
        {
          // In this case there are more than one user in the list.
          // Remove the user who requested the removal from the list:
          QStringList list = QStringList::split( ",", users, false );
          list.remove( getenv( "USER" ) );

          (*it).replace( users, list.join( "," ) );

          write = true;

          break;
        }
      }
      else
      {
        continue;
      }
    }

    if ( write )
    {
      // Prepare the contents: remove empty lines from the end.
      QStringList::Iterator it = contents.end();

      // Move the iterator to the last entry in the list:
      --it;

      while ( (*it).stripWhiteSpace().isEmpty() )
      {
        it = contents.remove( it );
        --it;
      }

      if ( !writeFile( filename, contents ) )
      {
        Smb4KError::error( ERROR_WRITING_FILE, filename );

        emit failed();
        emit finished();

        removeLockFile();
      }
      else
      {
        // Lock file file be removed at the end.

        emit finished();
      }
    }
    else
    {
      // Lock file will be removed at the end.

      emit finished();
    }
  }
  else if ( m_todo == SudoersRemove )
  {
    bool write = false;

    // Find the beginning and the end of the entries in
    // the super.tab file:
    QStringList::Iterator begin = contents.find( "# Entries for Smb4K users." );
    QStringList::Iterator end = contents.find( "# End of Smb4K user entries." );

    // Now, check if the user is in the list of users. If he is,
    // remove him from there or remove all if he is the only one:
    for ( QStringList::Iterator it = begin; it != end; ++it )
    {
      if ( (*it).startsWith( "User_Alias\tSMB4KUSERS" ) )
      {
        QString users = (*it).section( "=", 1, 1 ).stripWhiteSpace();

        if ( users.contains( "," ) == 0 )
        {
          // In this case, there is only one user in the list. Check if
          // it is the user who requested the removal:
          if ( QString::compare( users, getenv( "USER" ) ) == 0 )
          {
            // They are equal. Remove everything:
            contents.erase( begin, end );
            contents.remove( end );

            write = true;

            break;
          }
          else
          {
            // They are not equal: Do nothing.
            break;
          }
        }
        else
        {
          // In this case there are more than one user in the list.
          // Remove the user who requested the removal from the list:
          QStringList list = QStringList::split( ",", users, false );
          list.remove( getenv( "USER" ) );

          (*it).replace( users, list.join( "," ) );

          write = true;

          break;
        }
      }
      else
      {
        continue;
      }
    }

    if ( write )
    {
      // Prepare the contents: remove empty lines from the end.
      QStringList::Iterator it = contents.end();

      // Move the iterator to the last entry in the list:
      --it;

      while ( (*it).stripWhiteSpace().isEmpty() )
      {
        it = contents.remove( it );
        --it;
      }

      if ( !writeFile( filename, contents ) )
      {
        Smb4KError::error( ERROR_WRITING_FILE, filename );

        emit failed();
        emit finished();

        removeLockFile();
      }
      else
      {
        // Lock file file be removed at the end.

        emit finished();
      }
    }
    else
    {
      // Lock file will be removed at the end.

      emit finished();
    }
  }
  else
  {
    // Do nothing
  }

  removeLockFile();
}

/////////////////////////////////////////////////////////////////////////////
// SLOT IMPLEMENTATIONS
/////////////////////////////////////////////////////////////////////////////

void Smb4KFileIO::slotShutdown()
{
  removeLockFile( false );
}


void Smb4KFileIO::slotReceivedStderr( KProcess *, char *buf, int len )
{
  QString error_output = QString::fromLocal8Bit( buf, len );

  if ( error_output.contains( "smb4k_mv" ) != 0 )
  {
    m_error_code = Move;

    Smb4KError::error( ERROR_WRITING_FILE, (m_state == WriteSuperTab ? "super.tab" : "sudoers") );

    emit failed();
    emit finished();

    // We do not have a lock file, so we do not need to remove it!

    // FIXME: Do we need to write something to config() here? --- Test!
  }
  else if ( m_output_buffer.contains( "smb4k_cat" ) != 0 )
  {
    m_error_code = Cat;

    Smb4KError::error( ERROR_READING_FILE, (m_state == ReadSuperTab ? "super.tab" : "sudoers") );

    emit failed();
    emit finished();

    removeLockFile();
  }
}


void Smb4KFileIO::slotReceivedStdout( KProcess *, char *buf, int len )
{
  m_output_buffer.append( QString::fromLocal8Bit( buf, len ) );
}


void Smb4KFileIO::slotProcessExited( KProcess *p )
{
  // Delete the process:
  delete p;

  switch ( m_state )
  {
    case WriteSuperTab:
    case WriteSudoers:
    {
      // Error was already emitted by slotReceivedStderr().

      break;
    }
    case ReadSuperTab:
    {
      if ( m_error_code == None )
      {
        processOutput( "super.tab" );
      }
      else
      {
        // Error was already emitted by slotReceivedStderr().
        // Also the lock file has been removed.
      }
      break;
    }
    case ReadSudoers:
    {
      if ( m_error_code == None )
      {
        processOutput( "sudoers" );
      }
      else
      {
        // Error was already emitted by slotReceivedStderr().
        // Also the lock file has been removed.
      }
      break;
    }
    default:
    {
      break;
    }
  }

  m_state = Idle;
  m_todo = DoNothing;
  m_output_buffer = QString::null;
  m_error_code = None;
}

#include "smb4kfileio.moc"
